@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
-/***************************************************************************
+/*
+ * network-manager-openvpn - OpenVPN integration with NetworkManager
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,9 +15,9 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2008 - 2013 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.
- *
- **************************************************************************/
+ * Copyright (C) 2008 - 2013 Dan Williams <dcbw@redhat.com>
+ * Copyright (C) 2008 - 2018 Red Hat, Inc.
+ */
 
 #include "nm-default.h"
 
@@ -41,11 +41,13 @@
 #define INLINE_BLOB_CA                  NMV_OVPN_TAG_CA
 #define INLINE_BLOB_CERT                NMV_OVPN_TAG_CERT
 #define INLINE_BLOB_EXTRA_CERTS         NMV_OVPN_TAG_EXTRA_CERTS
+#define INLINE_BLOB_CRL_VERIFY          NMV_OVPN_TAG_CRL_VERIFY
 #define INLINE_BLOB_KEY                 NMV_OVPN_TAG_KEY
 #define INLINE_BLOB_PKCS12              NMV_OVPN_TAG_PKCS12
 #define INLINE_BLOB_SECRET              NMV_OVPN_TAG_SECRET
 #define INLINE_BLOB_TLS_AUTH            NMV_OVPN_TAG_TLS_AUTH
 #define INLINE_BLOB_TLS_CRYPT           NMV_OVPN_TAG_TLS_CRYPT
+#define INLINE_BLOB_TLS_CRYPT_V2        NMV_OVPN_TAG_TLS_CRYPT_V2
 
 const char *_nmovpn_test_temp_path = NULL;
 
@@ -775,8 +777,7 @@ inline_blob_write_out (const InlineBlobData *data, GError **error)
 NMConnection *
 do_import (const char *path, const char *contents, gsize contents_len, GError **error)
 {
-	gs_unref_object NMConnection *connection_free = NULL;
-	NMConnection *connection;
+	gs_unref_object NMConnection *connection = NULL;
 	NMSettingConnection *s_con;
 	NMSettingIPConfig *s_ip4;
 	NMSettingVpn *s_vpn;
@@ -791,13 +792,15 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 	char *tmp, *tmp2;
 	const char *ta_direction = NULL, *secret_direction = NULL;
 	gboolean allow_ta_direction = FALSE, allow_secret_direction = FALSE;
-	gboolean have_certs, have_ca;
-	GSList *inline_blobs = NULL, *sl_iter;
+	gboolean have_cert = FALSE, have_key = FALSE, have_ca = FALSE, have_pkcs12 = FALSE;
+	const char *cert_path = NULL, *key_path = NULL, *ca_path = NULL;
+	GSList *inline_blobs = NULL;
+	GSList *sl_iter;
 
 	g_return_val_if_fail (!error || !*error, NULL);
 
 	connection = nm_simple_connection_new ();
-	connection_free = connection;
+
 	s_con = NM_SETTING_CONNECTION (nm_setting_connection_new ());
 	nm_connection_add_setting (connection, NM_SETTING (s_con));
 	s_ip4 = NM_SETTING_IP_CONFIG (nm_setting_ip4_config_new ());
@@ -896,18 +899,17 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_PROTO)) {
 			if (!args_params_check_nargs_n (params, 1, &line_error))
 				goto handle_line_error;
-			/* Valid parameters are "udp", "tcp-client" and "tcp-server".
+			/* Valid parameters are defined in shared/utils.h
 			 * 'tcp' isn't technically valid, but it used to be accepted so
 			 * we'll handle it here anyway.
 			 */
-			if (nm_streq (params[1], "udp")) {
-				/* ignore; udp is default */
-			} else if (NM_IN_STRSET (params[1], "tcp-client", "tcp-server", "tcp"))
-				setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROTO_TCP, "yes");
-			else {
-				line_error = args_params_error_message_invalid_arg (params, 1);
+			if (!NM_IN_STRSET (params[1], NMOVPN_PROTCOL_TYPES)) {
+				line_error = g_strdup_printf (_("proto expects protocol type like “udp” or “tcp”"));
 				goto handle_line_error;
 			}
+			if (!NM_IN_STRSET (params[1], "udp", "udp4", "udp6")) {
+				setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROTO_TCP, "yes");
+			}
 			continue;
 		}
 
@@ -915,7 +917,7 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			if (!args_params_check_nargs_minmax (params, 0, 1, &line_error))
 				goto handle_line_error;
 			if (params[1]) {
-				if (!args_params_parse_int64 (params, 1, 1, G_MAXINT32, &v_int64, &line_error))
+				if (!args_params_parse_int64 (params, 1, 0, G_MAXINT32, &v_int64, &line_error))
 					goto handle_line_error;
 				setting_vpn_add_data_item_int64 (s_vpn, NM_OPENVPN_KEY_MSSFIX, v_int64);
 			} else
@@ -1007,7 +1009,7 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 
 			if (nm_streq (v, "no")) {
 				/* old plasma-nm used to set "comp-lzo=no" to mean unset, thus it spoiled
-				 * to "no" option to be used in the connection. Workaround that, by instead
+				 * the "no" option to be used in the connection. Workaround that, by instead
 				 * using "no-by-default" (bgo#769177). */
 				v = "no-by-default";
 			} else if (!NM_IN_STRSET (v, "yes", "adaptive")) {
@@ -1018,6 +1020,21 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			continue;
 		}
 
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_COMPRESS)) {
+			if (!args_params_check_nargs_minmax (params, 0, 1, &line_error))
+				goto handle_line_error;
+			if (params[1]) {
+				if (!NM_IN_STRSET (params[1], "lzo", "lz4", "lz4-v2")) {
+					line_error = g_strdup_printf (_("unsupported compress argument"));
+					goto handle_line_error;
+				}
+				setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_COMPRESS, params[1]);
+			} else {
+				setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_COMPRESS, "yes");
+			}
+			continue;
+		}
+
 		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_FLOAT)) {
 			if (!args_params_check_nargs_n (params, 0, &line_error))
 				goto handle_line_error;
@@ -1170,6 +1187,20 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			continue;
 		}
 
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_REMOTE_RANDOM_HOSTNAME)) {
+			if (!args_params_check_nargs_n (params, 0, &line_error))
+				goto handle_line_error;
+			setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME, "yes");
+			continue;
+		}
+
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_ALLOW_PULL_FQDN)) {
+			if (!args_params_check_nargs_n (params, 0, &line_error))
+				goto handle_line_error;
+			setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_ALLOW_PULL_FQDN, "yes");
+			continue;
+		}
+
 		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_TUN_IPV6)) {
 			if (!args_params_check_nargs_n (params, 0, &line_error))
 				goto handle_line_error;
@@ -1177,6 +1208,13 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			continue;
 		}
 
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_PUSH_PEER_INFO)) {
+			if (!args_params_check_nargs_n (params, 0, &line_error))
+				goto handle_line_error;
+			setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PUSH_PEER_INFO, "yes");
+			continue;
+		}
+
 		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_PORT, NMV_OVPN_TAG_RPORT)) {
 			if (!args_params_check_nargs_n (params, 1, &line_error))
 				goto handle_line_error;
@@ -1205,6 +1243,24 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			continue;
 		}
 
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_TLS_VERSION_MIN)) {
+			if (!args_params_check_nargs_n (params, 1, &line_error))
+				goto handle_line_error;
+			if (!args_params_check_arg_utf8 (params, 1, NULL, &line_error))
+				goto handle_line_error;
+			setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MIN, params[1]);
+			continue;
+		}
+
+		if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_TLS_VERSION_MAX)) {
+			if (!args_params_check_nargs_n (params, 1, &line_error))
+				goto handle_line_error;
+			if (!args_params_check_arg_utf8 (params, 1, NULL, &line_error))
+				goto handle_line_error;
+			setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MAX, params[1]);
+			continue;
+		}
+
 		if (NM_IN_STRSET (params[0],
 		                  NMV_OVPN_TAG_CA,
 		                  NMV_OVPN_TAG_CERT,
@@ -1213,7 +1269,8 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 		                  NMV_OVPN_TAG_PKCS12,
 		                  NMV_OVPN_TAG_SECRET,
 		                  NMV_OVPN_TAG_TLS_AUTH,
-		                  NMV_OVPN_TAG_TLS_CRYPT)) {
+		                  NMV_OVPN_TAG_TLS_CRYPT,
+		                  NMV_OVPN_TAG_TLS_CRYPT_V2)) {
 			const char *file;
 			gs_free char *file_free = NULL;
 			gboolean can_have_direction;
@@ -1239,18 +1296,24 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 				file = file_free = g_build_filename (default_path, file, NULL);
 
 			if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_PKCS12)) {
-				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_CA, file);
+				 /* OpenVPN allows --pkcs12 with external (PEM) --ca. Don't overwrite it with the PKCS#12 file. */
+				if (!have_ca) {
+					setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_CA, file);
+					have_ca = TRUE;
+				}
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_CERT, file);
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_KEY, file);
-			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_CA))
+				have_pkcs12 = TRUE;
+			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_CA)) {
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_CA, file);
-			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_EXTRA_CERTS))
-				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_EXTRA_CERTS, file);
-			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_CERT))
+				have_ca = TRUE;
+			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_CERT)) {
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_CERT, file);
-			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_KEY))
+				have_cert = TRUE;
+			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_KEY)) {
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_KEY, file);
-			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_SECRET)) {
+				have_key = TRUE;
+			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_SECRET)) {
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_STATIC_KEY, file);
 				if (s_direction)
 					secret_direction = s_direction;
@@ -1263,6 +1326,10 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 				allow_ta_direction = TRUE;
 			} else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_TLS_CRYPT))
 				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_TLS_CRYPT, file);
+			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_TLS_CRYPT_V2))
+				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_TLS_CRYPT_V2, file);
+			else if (NM_IN_STRSET (params[0], NMV_OVPN_TAG_EXTRA_CERTS))
+				setting_vpn_add_data_item_path (s_vpn, NM_OPENVPN_KEY_EXTRA_CERTS, file);
 			else
 				g_assert_not_reached ();
 			continue;
@@ -1424,7 +1491,7 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			}
 
 			{
-#ifdef NM_VPN_OLD
+#if ((NETWORKMANAGER_COMPILATION) & NM_NETWORKMANAGER_COMPILATION_WITH_LIBNM_UTIL)
 				NMIP4Route *route;
 
 				route = nm_ip4_route_new ();
@@ -1456,19 +1523,27 @@ do_import (const char *path, const char *contents, gsize contents_len, GError **
 			GString *blob_data;
 			InlineBlobData *inline_blob_data;
 
-			if (nm_streq (token, INLINE_BLOB_CA))
+			if (nm_streq (token, INLINE_BLOB_CA)) {
 				key = NM_OPENVPN_KEY_CA;
-			else if (nm_streq (token, INLINE_BLOB_CERT))
+				have_ca = TRUE;
+			} else if (nm_streq (token, INLINE_BLOB_CERT)) {
 				key = NM_OPENVPN_KEY_CERT;
-			else if (nm_streq (token, INLINE_BLOB_KEY))
+				have_cert = TRUE;
+			} else if (nm_streq (token, INLINE_BLOB_KEY)) {
 				key = NM_OPENVPN_KEY_KEY;
-			else if (nm_streq (token, INLINE_BLOB_EXTRA_CERTS))
-				key = NM_OPENVPN_KEY_EXTRA_CERTS;
-			else if (nm_streq (token, INLINE_BLOB_PKCS12)) {
+				have_key = TRUE;
+			} else if (nm_streq (token, INLINE_BLOB_PKCS12)) {
 				is_base64 = TRUE;
 				key = NULL;
-			} else if (nm_streq (token, INLINE_BLOB_TLS_CRYPT))
+				have_pkcs12 = TRUE;
+			} else if (nm_streq (token, INLINE_BLOB_EXTRA_CERTS))
+				key = NM_OPENVPN_KEY_EXTRA_CERTS;
+			else if (nm_streq (token, INLINE_BLOB_CRL_VERIFY))
+				key = NM_OPENVPN_KEY_CRL_VERIFY_FILE;
+			else if (nm_streq (token, INLINE_BLOB_TLS_CRYPT))
 				key = NM_OPENVPN_KEY_TLS_CRYPT;
+			else if (nm_streq (token, INLINE_BLOB_TLS_CRYPT_V2))
+				key = NM_OPENVPN_KEY_TLS_CRYPT_V2;
 			else if (nm_streq (token, INLINE_BLOB_TLS_AUTH)) {
 				key = NM_OPENVPN_KEY_TA;
 				allow_ta_direction = TRUE;
@@ -1585,28 +1660,77 @@ handle_line_error:
 		g_set_error_literal (error,
 		                     NMV_EDITOR_PLUGIN_ERROR,
 		                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
-		                     _("The file to import wasn’t a valid OpenVPN configure (no remote)"));
+		                     _("The file to import wasn’t a valid OpenVPN configuration (no remote)"));
 		goto out_error;
 	}
 
-	have_certs = FALSE;
-	have_ca = FALSE;
+	/* Validate PKCS#12/PEM/CA settings. PKCS#12 and PEM cannot be mixed,
+	 * with the exception of PKCS#12 cert + PEM CA. */
+	if (have_client && have_pkcs12) {
+		if (have_cert || have_key) {
+			g_set_error_literal (error,
+			                     NMV_EDITOR_PLUGIN_ERROR,
+			                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+			                     _("The file to import wasn’t a valid OpenVPN configuration (--pkcs12 can not be used with --cert or --key)"));
+			goto out_error;
+		}
 
-	if (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CA))
-		have_ca = TRUE;
+		if (have_ca) {
+			ca_path = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CA);
+			if (is_pkcs12 (ca_path)) {
+				g_set_error_literal (error,
+				                     NMV_EDITOR_PLUGIN_ERROR,
+				                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+				                     _("The file to import wasn’t a valid OpenVPN configuration (--ca can not be PKCS#12 format)"));
+				goto out_error;
+			}
+		}
+	} else if (have_client && !have_pkcs12) {
+		if (!have_ca) {
+			g_set_error_literal (error,
+			                     NMV_EDITOR_PLUGIN_ERROR,
+			                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+			                     _("The file to import wasn’t a valid OpenVPN configuration (missing --ca)"));
+			goto out_error;
+		}
 
-	if (   have_ca
-	    && nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CERT)
-	    && nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEY))
-		have_certs = TRUE;
+		ca_path = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CA);
+
+		if ((have_cert || have_key) && !(have_cert && have_key)) {
+			g_set_error_literal (error,
+			                     NMV_EDITOR_PLUGIN_ERROR,
+			                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+			                     _("The file to import wasn’t a valid OpenVPN configuration (if you use one of --cert or --key, you must use them both)"));
+			goto out_error;
+		}
+
+		cert_path = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CERT);
+		key_path = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEY);
+
+		if (cert_path && is_pkcs12 (cert_path)) {
+			g_set_error_literal (error,
+			                     NMV_EDITOR_PLUGIN_ERROR,
+			                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+			                     _("The file to import wasn’t a valid OpenVPN configuration (--cert can not be PKCS#12 format.)"));
+			goto out_error;
+		}
+
+		if (key_path && is_pkcs12 (key_path)) {
+			g_set_error_literal (error,
+			                     NMV_EDITOR_PLUGIN_ERROR,
+			                     NMV_EDITOR_PLUGIN_ERROR_FILE_NOT_VPN,
+			                     _("The file to import wasn’t a valid OpenVPN configuration (--key can not be PKCS#12 format.)"));
+			goto out_error;
+		}
+	}
 
 	/* Determine connection type */
 	if (have_pass) {
-		if (have_certs)
+		if (have_cert || have_pkcs12)
 			ctype = NM_OPENVPN_CONTYPE_PASSWORD_TLS;
 		else if (have_ca)
 			ctype = NM_OPENVPN_CONTYPE_PASSWORD;
-	} else if (have_certs) {
+	} else if (have_cert || have_pkcs12) {
 		ctype = NM_OPENVPN_CONTYPE_TLS;
 	} else if (have_sk)
 		ctype = NM_OPENVPN_CONTYPE_STATIC_KEY;
@@ -1623,15 +1747,12 @@ handle_line_error:
 		                             NM_SETTING_SECRET_FLAG_AGENT_OWNED,
 		                             NULL);
 	}
-	if (have_certs) {
+	if (have_key || have_pkcs12) {
 		gs_free char *key_path_free = NULL;
-		const char *key_path;
 
-		key_path = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEY);
 		if (is_encrypted (nm_utils_str_utf8safe_unescape (key_path, &key_path_free))) {
 			/* If there should be a private key password, default it to
-			 * being agent-owned.
-			 */
+			 * being agent-owned. */
 			nm_setting_set_secret_flags (NM_SETTING (s_vpn),
 			                             NM_OPENVPN_KEY_CERTPASS,
 			                             NM_SETTING_SECRET_FLAG_AGENT_OWNED,
@@ -1639,28 +1760,50 @@ handle_line_error:
 		}
 	}
 
-	inline_blobs = g_slist_reverse (inline_blobs);
-	for (sl_iter = inline_blobs; sl_iter; sl_iter = sl_iter->next) {
-		const InlineBlobData *data = sl_iter->data;
+	if (inline_blobs) {
+		GSList *tmp_list = NULL;
 
-		/* Check whether the setting was not overwritten by a later entry in the config-file. */
-		if (nm_streq (data->token, INLINE_BLOB_PKCS12)) {
-			if (   !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_CA, data->path)
-			    && !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_CERT, data->path)
-			    && !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_KEY, data->path))
-				continue;
-		} else {
-			if (!setting_vpn_eq_data_item_utf8safe (s_vpn, data->key, data->path))
+		/* filter out blobs that are shadowed and not used. */
+		while ((sl_iter = inline_blobs)) {
+			InlineBlobData *data = sl_iter->data;
+			gboolean is_good = TRUE;
+
+			inline_blobs = inline_blobs->next;
+
+			/* Check whether the setting was not overwritten by a later entry in the config-file. */
+			if (nm_streq (data->token, INLINE_BLOB_PKCS12)) {
+				if (   !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_CA, data->path)
+				    && !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_CERT, data->path)
+				    && !setting_vpn_eq_data_item_utf8safe (s_vpn, NM_OPENVPN_KEY_KEY, data->path))
+					is_good = FALSE;
+			} else {
+				if (!setting_vpn_eq_data_item_utf8safe (s_vpn, data->key, data->path))
+					is_good = FALSE;
+			}
+			if (!is_good) {
+				g_slist_free_1 (sl_iter);
+				inline_blob_data_free (data);
 				continue;
+			}
+
+			/* we prepend the element to a new, temporary list. That means, we
+			 * reverse the order here. That is just what we need, because the
+			 * list was previously constructed in reverse order. */
+			sl_iter->next = tmp_list;
+			tmp_list = sl_iter;
 		}
+		inline_blobs = tmp_list;
+	}
+
+	for (sl_iter = inline_blobs; sl_iter; sl_iter = sl_iter->next) {
 		if (!inline_blob_write_out (sl_iter->data, error))
 			goto out_error;
 	}
+
 	g_slist_free_full (inline_blobs, (GDestroyNotify) inline_blob_data_free);
 
-	connection_free = NULL;
 	g_return_val_if_fail (!error || !*error, connection);
-	return connection;
+	return g_steal_pointer (&connection);
 
 out_error:
 	g_slist_free_full (inline_blobs, (GDestroyNotify) inline_blob_data_free);
@@ -1675,7 +1818,7 @@ escape_arg (const char *value, char **buf)
 {
 	const char *s;
 	char *result, *i_result;
-	gboolean has_single_quote = FALSE;
+	gboolean needs_double_quotes = FALSE;
 	gboolean needs_quotation = FALSE;
 	gsize len;
 
@@ -1696,14 +1839,15 @@ escape_arg (const char *value, char **buf)
 		    || (c >= 'A' && c <= 'Z')
 		    || NM_IN_SET (c, '_', '-', ':', '/'))
 			continue;
+
 		needs_quotation = TRUE;
-		if (c == '\'')
-			has_single_quote = TRUE;
+		if (NM_IN_SET (c, '\'', '\n'))
+			needs_double_quotes = TRUE;
 	}
 	if (!needs_quotation)
 		return value;
 
-	if (!has_single_quote) {
+	if (!needs_double_quotes) {
 		result = g_malloc (len + 2 + 1);
 		result[0] = '\'';
 		memcpy (&result[1], value, len);
@@ -1713,9 +1857,19 @@ escape_arg (const char *value, char **buf)
 		i_result = result = g_malloc (len * 2 + 3);
 		*(i_result++) = '"';
 		for (s = value; s[0]; s++) {
-			if (NM_IN_SET (s[0], '\\', '"'))
+			if (s[0] == '\n') {
+				/* Openvpn does not support encoding '\n' in the ovpn configuration file.
+				 * This configuration cannot be expressed in an ovpn file.
+				 *
+				 * Still escape it as '\n', although openvpn's parser will warn
+				 * about the invalid escape sequence. */
 				*(i_result++) = '\\';
-			*(i_result++) = s[0];
+				*(i_result++) = 'n';
+			} else {
+				if (NM_IN_SET (s[0], '\\', '"'))
+					*(i_result++) = '\\';
+				*(i_result++) = s[0];
+			}
 		}
 		*(i_result++) = '"';
 		*(i_result++) = '\0';
@@ -1844,7 +1998,10 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 
 	f = g_string_sized_new (512);
 
-	args_write_line (f, NMV_OVPN_TAG_CLIENT);
+	if (NM_IN_STRSET (connection_type, NM_OPENVPN_CONTYPE_TLS,
+	                                   NM_OPENVPN_CONTYPE_PASSWORD,
+	                                   NM_OPENVPN_CONTYPE_PASSWORD_TLS))
+		args_write_line (f, NMV_OVPN_TAG_CLIENT);
 
 	/* 'remote' */
 	gw_list = g_strsplit_set (gateways, " \t,", 0);
@@ -1872,9 +2029,18 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_RANDOM), "yes"))
 		args_write_line (f, NMV_OVPN_TAG_REMOTE_RANDOM);
 
+	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME), "yes"))
+		args_write_line (f, NMV_OVPN_TAG_REMOTE_RANDOM_HOSTNAME);
+
+	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_ALLOW_PULL_FQDN), "yes"))
+		args_write_line (f, NMV_OVPN_TAG_ALLOW_PULL_FQDN);
+
 	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TUN_IPV6), "yes"))
 		args_write_line (f, NMV_OVPN_TAG_TUN_IPV6);
 
+	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PUSH_PEER_INFO), "yes"))
+		args_write_line (f, NMV_OVPN_TAG_PUSH_PEER_INFO);
+
 	{
 		gs_free char *cacert_free = NULL, *user_cert_free = NULL, *private_key_free = NULL;
 		const char *cacert = NULL, *user_cert = NULL, *private_key = NULL;
@@ -1898,18 +2064,21 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 				private_key = nm_utils_str_utf8safe_unescape (value, &private_key_free);
 		}
 
-		if (   cacert && user_cert && private_key
-		    && nm_streq (cacert, user_cert) && nm_streq (cacert, private_key)) {
-			/* Handle PKCS#12 (all certs are the same file) */
-			args_write_line (f, NMV_OVPN_TAG_PKCS12, cacert);
-		} else {
-			if (cacert)
-				args_write_line (f, NMV_OVPN_TAG_CA, cacert);
-			if (user_cert)
+		if (user_cert && private_key) {
+			if (   nm_streq (user_cert, private_key)
+			    && is_pkcs12 (user_cert))
+				args_write_line (f, NMV_OVPN_TAG_PKCS12, user_cert);
+			else {
 				args_write_line (f, NMV_OVPN_TAG_CERT, user_cert);
-			if (private_key)
 				args_write_line (f, NMV_OVPN_TAG_KEY, private_key);
+			}
 		}
+
+		/* The CA can be included in a PKCS#12 file above, or a separate file (PEM format).
+		 * If it's pkcs12 we can't use --ca as it only accepts PEM format. */
+		if (   cacert
+		    && !is_pkcs12 (cacert))
+			args_write_line (f, NMV_OVPN_TAG_CA, cacert);
 	}
 
 	if (NM_IN_STRSET (connection_type, NM_OPENVPN_CONTYPE_PASSWORD,
@@ -1945,6 +2114,12 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 		args_write_line (f, NMV_OVPN_TAG_COMP_LZO, value);
 	}
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMPRESS);
+	if (nm_streq0 (value, "yes"))
+		args_write_line (f, NMV_OVPN_TAG_COMPRESS);
+	else if (value)
+		args_write_line_setting_value (f, NMV_OVPN_TAG_COMPRESS, s_vpn, NM_OPENVPN_KEY_COMPRESS);
+
 	if (nm_streq0 (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_FLOAT), "yes"))
 		args_write_line (f, NMV_OVPN_TAG_FLOAT);
 
@@ -2047,6 +2222,22 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 			                 nm_utils_str_utf8safe_unescape (key, &s_free));
 		}
 
+		key = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_CRYPT_V2);
+		if (nmovpn_arg_is_set (key)) {
+			gs_free char *s_free = NULL;
+			args_write_line (f,
+			                 NMV_OVPN_TAG_TLS_CRYPT_V2,
+			                 nm_utils_str_utf8safe_unescape (key, &s_free));
+		}
+
+		key = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MIN);
+		if (nmovpn_arg_is_set (key))
+			args_write_line (f, NMV_OVPN_TAG_TLS_VERSION_MIN, key);
+
+		key = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MAX);
+		if (nmovpn_arg_is_set (key))
+			args_write_line (f, NMV_OVPN_TAG_TLS_VERSION_MAX, key);
+
 		key = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_EXTRA_CERTS);
 		if (nmovpn_arg_is_set (key)) {
 			gs_free char *s_free = NULL;
@@ -2111,7 +2302,7 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 
 	s_ip4 = nm_connection_get_setting_ip4_config (connection);
 	if (s_ip4) {
-#ifdef NM_VPN_OLD
+#if ((NETWORKMANAGER_COMPILATION) & NM_NETWORKMANAGER_COMPILATION_WITH_LIBNM_UTIL)
 		num = nm_setting_ip4_config_get_num_routes (s_ip4);
 #else
 		num = nm_setting_ip_config_get_num_routes (s_ip4);
@@ -2124,7 +2315,7 @@ do_export_create (NMConnection *connection, const char *path, GError **error)
 			guint64 metric;
 			char metric_buf[50];
 
-#ifdef NM_VPN_OLD
+#if ((NETWORKMANAGER_COMPILATION) & NM_NETWORKMANAGER_COMPILATION_WITH_LIBNM_UTIL)
 			char next_hop_str_buf[INET_ADDRSTRLEN] = { 0 };
 			char dest_str_buf[INET_ADDRSTRLEN] = { 0 };
 			in_addr_t dest, next_hop;
