@@ -1,7 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
-/***************************************************************************
- *
- * Copyright (C) 2008 Dan Williams, <dcbw@redhat.com>
+/*
+ * network-manager-openvpn - OpenVPN integration with NetworkManager
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,20 +15,21 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- **************************************************************************/
+ * Copyright (C) 2008 Dan Williams, <dcbw@redhat.com>
+ * Copyright (C) 2008 - 2018 Red Hat, Inc.
+ */
 
 #ifndef _IMPORT_EXPORT_H_
 #define _IMPORT_EXPORT_H_
 
-#include <glib.h>
+extern const char *_nmovpn_test_temp_path;
 
-#ifdef NM_OPENVPN_OLD
-#include <nm-connection.h>
-#else /* !NM_OPENVPN_OLD */
-#include <NetworkManager.h>
-#endif
+gboolean _nmovpn_test_args_parse_line (const char *line,
+                                       gsize line_len,
+                                       const char ***out_p,
+                                       char **out_error);
 
-NMConnection *do_import (const char *path, char **lines, GError **error);
+NMConnection *do_import (const char *path, const char *contents, gsize contents_len, GError **error);
 
 gboolean do_export (const char *path, NMConnection *connection, GError **error);
 
