@@ -1,14 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
-/***************************************************************************
- * CVSID: $Id: nm-openvpn.c 4232 2008-10-29 09:13:40Z tambeti $
- *
- * nm-openvpn.c : GNOME UI dialogs for configuring openvpn VPN connections
- *
- * Copyright (C) 2005 Tim Niemueller <tim@niemueller.de>
- * Copyright (C) 2008 Tambet Ingo, <tambet@gmail.com>
- * Copyright (C) 2008 - 2010 Dan Williams, <dcbw@redhat.com>
- * Copyright (C) 2008 - 2017 Red Hat, Inc.
- * Based on work by David Zeuthen, <davidz@redhat.com>
+/*
+ * network-manager-openvpn - OpenVPN integration with NetworkManager
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,7 +15,12 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- **************************************************************************/
+ * Based on work by David Zeuthen, <davidz@redhat.com>
+ * Copyright (C) 2005 Tim Niemueller <tim@niemueller.de>
+ * Copyright (C) 2008 Tambet Ingo, <tambet@gmail.com>
+ * Copyright (C) 2008 - 2010 Dan Williams, <dcbw@redhat.com>
+ * Copyright (C) 2008 - 2017 Red Hat, Inc.
+ */
 
 #include "nm-default.h"
 
@@ -56,38 +52,6 @@ static GtkFileFilter *sk_file_chooser_filter_new (void);
 
 /*****************************************************************************/
 
-static const char *comp_lzo_values[] = {
-	"adaptive",
-	"yes",
-	"no-by-default",
-};
-
-static const char *
-comp_lzo_values_conf_coerce (const char *value_conf)
-{
-	if (!value_conf || nm_streq (value_conf, "no"))
-		return NULL;
-	if (nm_streq (value_conf, "yes"))
-		return "yes";
-	if (nm_streq (value_conf, "no-by-default"))
-		return "no-by-default";
-	return "adaptive";
-}
-
-static const char *
-comp_lzo_values_conf_to_display (const char *value_conf)
-{
-	if (nm_streq (value_conf, "yes"))
-		return "yes";
-	if (nm_streq (value_conf, "no-by-default"))
-		return "no";
-	if (nm_streq (value_conf, "adaptive"))
-		return "adaptive";
-	g_return_val_if_reached ("adaptive");
-}
-
-/*****************************************************************************/
-
 /* From gnome-control-center/panels/network/connection-editor/ui-helpers.c */
 
 static void
@@ -108,29 +72,58 @@ widget_unset_error (GtkWidget *widget)
 
 /*****************************************************************************/
 
+static void
+tls_ca_changed_cb (NMACertChooser *this, gpointer user_data)
+{
+	NMACertChooser *other = user_data;
+	NMSetting8021xCKScheme scheme;
+	gs_free char *ca_cert = NULL;
+	gs_free char *client_cert = NULL;
+	gs_free char *client_key = NULL;
+
+	client_key = nma_cert_chooser_get_key (other, &scheme);
+	client_cert = nma_cert_chooser_get_cert (other, &scheme);
+	ca_cert = nma_cert_chooser_get_cert (this, &scheme);
+
+	/* OpenVPN allows --pkcs12 with --ca, but if the provided CA is a PKCS#12 file,
+	 * we have to also set the cert/key to the same file - since OpenVPN does not
+	 * allow mixing of --pkcs12 and --cert/--key. */
+	if (   scheme == NM_SETTING_802_1X_CK_SCHEME_PATH
+	    && is_pkcs12 (ca_cert)) {
+		nma_cert_chooser_set_cert (other, ca_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
+		nma_cert_chooser_set_key (other, ca_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
+	}
+}
+
 static void
 tls_cert_changed_cb (NMACertChooser *this, gpointer user_data)
 {
 	NMACertChooser *other = user_data;
 	NMSetting8021xCKScheme scheme;
-	gs_free char *this_cert = NULL;
-	gs_free char *other_cert = NULL;
-	gs_free char *this_key = NULL;
-	gs_free char *other_key = NULL;
+	gs_free char *ca_cert = NULL;
+	gs_free char *client_cert = NULL;
+	gs_free char *client_key = NULL;
 
-	other_key = nma_cert_chooser_get_key (other, &scheme);
-	this_key = nma_cert_chooser_get_key (this, &scheme);
-	other_cert = nma_cert_chooser_get_cert (other, &scheme);
-	this_cert = nma_cert_chooser_get_cert (this, &scheme);
-	if (   scheme == NM_SETTING_802_1X_CK_SCHEME_PATH
-	    && is_pkcs12 (this_cert)) {
-		if (!this_key)
-			nma_cert_chooser_set_key (this, this_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
-		if (!other_cert) {
-			nma_cert_chooser_set_cert (other, this_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
-			if (!other_key)
-				nma_cert_chooser_set_key (other, this_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
+	ca_cert = nma_cert_chooser_get_cert (other, &scheme);
+	client_key = nma_cert_chooser_get_key (this, &scheme);
+	client_cert = nma_cert_chooser_get_cert (this, &scheme);
+
+	/* OpenVPN does not allow a combination of --cert/--key and --pkcs12; however,
+	 * it does allow --pkcs12 with --ca. */
+	if (client_cert && is_pkcs12 (client_cert)) {
+		if (!ca_cert)
+			nma_cert_chooser_set_cert (other, client_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
+		if (   ca_cert
+		    && is_pkcs12 (ca_cert)
+		    && !nm_streq (client_cert, ca_cert))
+			nma_cert_chooser_set_cert (other, client_cert, NM_SETTING_802_1X_CK_SCHEME_PATH);
+	} else if (client_cert && !is_pkcs12 (client_cert)) {
+		if (client_key && is_pkcs12 (client_key)) {
+			nma_cert_chooser_set_key (this, NULL, NM_SETTING_802_1X_CK_SCHEME_UNKNOWN);
+			nma_cert_chooser_set_cert_password (this, "");
 		}
+		if (ca_cert && is_pkcs12 (ca_cert))
+			nma_cert_chooser_set_cert (other, NULL, NM_SETTING_802_1X_CK_SCHEME_UNKNOWN);
 	}
 }
 
@@ -168,8 +161,8 @@ tls_setup (GtkBuilder *builder,
 	nma_cert_chooser_setup_key_password_storage (cert, 0, (NMSetting *) s_vpn,
 	                                             NM_OPENVPN_KEY_CERTPASS, TRUE, FALSE);
 
-	/* Link choosers to the PKCS#12 changer callback */
-	g_signal_connect_object (ca_chooser, "changed", G_CALLBACK (tls_cert_changed_cb), cert, 0);
+	/* Link choosers to the PKCS#12 changer callbacks */
+	g_signal_connect_object (ca_chooser, "changed", G_CALLBACK (tls_ca_changed_cb), cert, 0);
 	g_signal_connect_object (cert, "changed", G_CALLBACK (tls_cert_changed_cb), ca_chooser, 0);
 }
 
@@ -682,8 +675,10 @@ sk_file_chooser_filter_new (void)
 }
 
 static const char *const advanced_keys[] = {
+	NM_OPENVPN_KEY_ALLOW_PULL_FQDN,
 	NM_OPENVPN_KEY_AUTH,
 	NM_OPENVPN_KEY_CIPHER,
+	NM_OPENVPN_KEY_COMPRESS,
 	NM_OPENVPN_KEY_COMP_LZO,
 	NM_OPENVPN_KEY_CONNECT_TIMEOUT,
 	NM_OPENVPN_KEY_CRL_VERIFY_DIR,
@@ -698,6 +693,7 @@ static const char *const advanced_keys[] = {
 	NM_OPENVPN_KEY_MAX_ROUTES,
 	NM_OPENVPN_KEY_MSSFIX,
 	NM_OPENVPN_KEY_MTU_DISC,
+	NM_OPENVPN_KEY_NCP_DISABLE,
 	NM_OPENVPN_KEY_NS_CERT_TYPE,
 	NM_OPENVPN_KEY_PING,
 	NM_OPENVPN_KEY_PING_EXIT,
@@ -708,14 +704,19 @@ static const char *const advanced_keys[] = {
 	NM_OPENVPN_KEY_PROXY_RETRY,
 	NM_OPENVPN_KEY_PROXY_SERVER,
 	NM_OPENVPN_KEY_PROXY_TYPE,
+	NM_OPENVPN_KEY_PUSH_PEER_INFO,
 	NM_OPENVPN_KEY_REMOTE_CERT_TLS,
 	NM_OPENVPN_KEY_REMOTE_RANDOM,
+	NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME,
 	NM_OPENVPN_KEY_RENEG_SECONDS,
 	NM_OPENVPN_KEY_TA,
 	NM_OPENVPN_KEY_TAP_DEV,
 	NM_OPENVPN_KEY_TA_DIR,
 	NM_OPENVPN_KEY_TLS_CRYPT,
+	NM_OPENVPN_KEY_TLS_CRYPT_V2,
 	NM_OPENVPN_KEY_TLS_REMOTE,
+	NM_OPENVPN_KEY_TLS_VERSION_MIN,
+	NM_OPENVPN_KEY_TLS_VERSION_MAX,
 	NM_OPENVPN_KEY_TUNNEL_MTU,
 	NM_OPENVPN_KEY_TUN_IPV6,
 	NM_OPENVPN_KEY_VERIFY_X509_NAME,
@@ -844,8 +845,7 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 
 		/* Don't add anything until after the first blank line. Also,
 		 * any blank line indicates the start of a comment, ended by
-		 * another blank line.
-		 */
+		 * another blank line. */
 		if (!strlen (*item)) {
 			ignore_lines = !ignore_lines;
 			continue;
@@ -1122,9 +1122,10 @@ populate_remote_cert_tls_combo (GtkComboBox *box, const char *remote_cert)
 	g_object_unref (store);
 }
 
-#define TLS_AUTH_MODE_NONE  0
-#define TLS_AUTH_MODE_AUTH  1
-#define TLS_AUTH_MODE_CRYPT 2
+#define TLS_AUTH_MODE_NONE     0
+#define TLS_AUTH_MODE_AUTH     1
+#define TLS_AUTH_MODE_CRYPT    2
+#define TLS_AUTH_MODE_CRYPT_V2 3
 
 static void
 tls_auth_toggled_cb (GtkWidget *widget, gpointer user_data)
@@ -1253,8 +1254,7 @@ proxy_type_changed (GtkComboBox *combo, gpointer user_data)
 	}
 
 	/* Proxy options require TCP; but don't reset the TCP checkbutton
-	 * to false when the user disables HTTP proxy; leave it checked.
-	 */
+	 * to false when the user disables HTTP proxy; leave it checked. */
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tcp_checkbutton"));
 	if (sensitive == TRUE)
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
@@ -1457,15 +1457,15 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	GtkBuilder *builder;
 	GtkWidget *dialog = NULL;
 	GtkWidget *widget, *combo, *spin, *entry, *ok_button;
-	const char *value, *value2;
+	const char *value, *value2, *value3;
 	const char *dev, *dev_type, *tap_dev;
 	GtkListStore *store;
 	GtkTreeIter iter;
 	int vint;
-	guint i;
 	guint32 active;
 	NMSettingSecretFlags pw_flags;
 	GError *error = NULL;
+	NMOvpnComp comp;
 
 	g_return_val_if_fail (hash != NULL, NULL);
 
@@ -1583,29 +1583,19 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	_builder_init_optional_spinbutton (builder, "fragment_checkbutton", "fragment_spinbutton", !!value,
 	                                   _nm_utils_ascii_str_to_int64 (value, 10, 0, 65535, 1300));
 
+	comp = nmovpn_compression_from_options (g_hash_table_lookup (hash, NM_OPENVPN_KEY_COMP_LZO),
+	                                        g_hash_table_lookup (hash, NM_OPENVPN_KEY_COMPRESS));
 
-	value = comp_lzo_values_conf_coerce (g_hash_table_lookup (hash, NM_OPENVPN_KEY_COMP_LZO));
-	widget = GTK_WIDGET (_builder_init_toggle_button (builder, "lzo_checkbutton", value != NULL));
-	combo = GTK_WIDGET (gtk_builder_get_object (builder, "lzo_combo"));
-	store = gtk_list_store_new (1, G_TYPE_STRING);
-	active = 0;
-	for (i = 0; i < G_N_ELEMENTS (comp_lzo_values); i++) {
-		gtk_list_store_append (store, &iter);
-		gtk_list_store_set (store, &iter,
-		                    0, comp_lzo_values_conf_to_display (comp_lzo_values[i]),
-		                    -1);
-		if (nm_streq (comp_lzo_values[i], value ?: "adaptive"))
-			active = i;
-	}
-	gtk_combo_box_set_model (GTK_COMBO_BOX (combo), GTK_TREE_MODEL (store));
-	g_object_unref (store);
-	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), active);
+	combo = GTK_WIDGET (gtk_builder_get_object (builder, "compress_combo"));
+	widget = GTK_WIDGET (_builder_init_toggle_button (builder, "compress_checkbutton", comp != NMOVPN_COMP_DISABLED));
 	g_object_bind_property (widget, "active", combo, "sensitive", G_BINDING_SYNC_CREATE);
+	if (comp != NMOVPN_COMP_DISABLED)
+		gtk_combo_box_set_active (GTK_COMBO_BOX (combo), comp - 1);
 
 	_builder_init_toggle_button (builder, "mssfix_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_MSSFIX));
 	_builder_init_toggle_button (builder, "float_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_FLOAT));
 	_builder_init_toggle_button (builder, "tcp_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_PROTO_TCP));
-
+	_builder_init_toggle_button (builder, "ncp_disable_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_NCP_DISABLE));
 
 	/* Populate device-related widgets */
 	dev =      g_hash_table_lookup (hash, NM_OPENVPN_KEY_DEV);
@@ -1641,6 +1631,8 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 
 
 	_builder_init_toggle_button (builder, "remote_random_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_REMOTE_RANDOM));
+	_builder_init_toggle_button (builder, "remote_random_hostname_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME));
+	_builder_init_toggle_button (builder, "allow_pull_fqdn_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_ALLOW_PULL_FQDN));
 	_builder_init_toggle_button (builder, "tun_ipv6_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_TUN_IPV6));
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "cipher_combo"));
@@ -1709,7 +1701,11 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_auth_chooser"));
 		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TA);
 		value2 = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_CRYPT);
-		if (value2 && value2[0]) {
+		value3 = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_CRYPT_V2);
+		if (value3 && value3[0]) {
+			gtk_combo_box_set_active (GTK_COMBO_BOX (combo), TLS_AUTH_MODE_CRYPT_V2);
+			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value3);
+		} else if (value2 && value2[0]) {
 			gtk_combo_box_set_active (GTK_COMBO_BOX (combo), TLS_AUTH_MODE_CRYPT);
 			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value2);
 		} else if (value && value[0]) {
@@ -1814,6 +1810,20 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	_builder_init_optional_spinbutton (builder, "max_routes_checkbutton", "max_routes_spinbutton", !!value,
 	                                   _nm_utils_ascii_str_to_int64 (value, 10, 0, 100000000, 100));
 
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_VERSION_MIN);
+	if (value && *value) {
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_version_min"));
+		gtk_entry_set_text (GTK_ENTRY (widget), value);
+	}
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_VERSION_MAX);
+	if (value && *value) {
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_version_max"));
+		gtk_entry_set_text (GTK_ENTRY (widget), value);
+	}
+
+	_builder_init_toggle_button (builder, "push_peer_info_checkbutton",
+	                             _hash_get_boolean (hash, NM_OPENVPN_KEY_PUSH_PEER_INFO));
+
 	return dialog;
 }
 
@@ -1825,7 +1835,6 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 	GtkBuilder *builder;
 	const char *contype = NULL;
 	const char *value;
-	int active;
 	int proxy_type = PROXY_TYPE_NONE;
 	GtkTreeModel *model;
 	GtkTreeIter iter;
@@ -1932,12 +1941,19 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 		}
 	}
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "lzo_checkbutton"));
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "compress_checkbutton"));
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
-		combo = GTK_WIDGET (gtk_builder_get_object (builder, "lzo_combo"));
-		active = gtk_combo_box_get_active (GTK_COMBO_BOX (combo));
-		if (active >= 0 && active < G_N_ELEMENTS (comp_lzo_values))
-			g_hash_table_insert (hash, NM_OPENVPN_KEY_COMP_LZO, g_strdup (comp_lzo_values[active]));
+		const char *opt_compress;
+		const char *opt_comp_lzo;
+		NMOvpnComp comp;
+
+		combo = GTK_WIDGET (gtk_builder_get_object (builder, "compress_combo"));
+		comp = gtk_combo_box_get_active (GTK_COMBO_BOX (combo)) + 1;
+		nmovpn_compression_to_options (comp, &opt_comp_lzo, &opt_compress);
+		if (opt_compress)
+			g_hash_table_insert (hash, NM_OPENVPN_KEY_COMPRESS, g_strdup (opt_compress));
+		if (opt_comp_lzo)
+			g_hash_table_insert (hash, NM_OPENVPN_KEY_COMP_LZO, g_strdup (opt_comp_lzo));
 	}
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "mssfix_checkbutton"));
@@ -1952,6 +1968,10 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, NM_OPENVPN_KEY_PROTO_TCP, g_strdup ("yes"));
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "ncp_disable_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_NCP_DISABLE, g_strdup ("yes"));
+
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "dev_checkbutton"));
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
 		int device_type;
@@ -1972,6 +1992,14 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, NM_OPENVPN_KEY_REMOTE_RANDOM, g_strdup ("yes"));
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_random_hostname_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME, g_strdup ("yes"));
+
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "allow_pull_fqdn_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_ALLOW_PULL_FQDN, g_strdup ("yes"));
+
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tun_ipv6_checkbutton"));
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, NM_OPENVPN_KEY_TUN_IPV6, g_strdup ("yes"));
@@ -2011,6 +2039,15 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 		if (hmacauth)
 			g_hash_table_insert (hash, NM_OPENVPN_KEY_AUTH, hmacauth);
 	}
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, "tls_version_min"));
+	value = gtk_entry_get_text (GTK_ENTRY (entry));
+	if (value && *value)
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_TLS_VERSION_MIN, g_strdup (value));
+
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, "tls_version_max"));
+	value = gtk_entry_get_text (GTK_ENTRY (entry));
+	if (value && *value)
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_TLS_VERSION_MAX, g_strdup (value));
 
 	contype = g_object_get_data (G_OBJECT (dialog), "connection-type");
 	if (   !strcmp (contype, NM_OPENVPN_CONTYPE_TLS)
@@ -2100,6 +2137,13 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 				g_hash_table_insert (hash, NM_OPENVPN_KEY_TLS_CRYPT, g_strdup (filename));
 			g_free (filename);
 			break;
+		case TLS_AUTH_MODE_CRYPT_V2:
+			widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_auth_chooser"));
+			filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
+			if (filename && filename[0])
+				g_hash_table_insert (hash, NM_OPENVPN_KEY_TLS_CRYPT_V2, g_strdup (filename));
+			g_free (filename);
+			break;
 		case TLS_AUTH_MODE_NONE:
 			break;
 		}
@@ -2195,6 +2239,10 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog)
 		}
 	}
 
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "push_peer_info_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		g_hash_table_insert (hash, NM_OPENVPN_KEY_PUSH_PEER_INFO, g_strdup ("yes"));
+
 	return hash;
 }
 
@@ -2497,7 +2545,7 @@ hash_copy_advanced (gpointer key, gpointer data, gpointer user_data)
 	g_return_if_fail (value && *value);
 
 	/* HTTP Proxy password is a secret, not a data item */
-	if (NM_IN_SET (key, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD))
+	if (nm_streq0 (key, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD))
 		nm_setting_vpn_add_secret (s_vpn, (const char *) key, value);
 	else
 		nm_setting_vpn_add_data_item (s_vpn, (const char *) key, value);
@@ -2690,6 +2738,7 @@ nm_vpn_editor_factory_openvpn (NMVpnEditorPlugin *editor_plugin,
                                NMConnection *connection,
                                GError **error)
 {
+	g_type_ensure (NMA_TYPE_CERT_CHOOSER);
 	return openvpn_editor_new (connection, error);
 }
 #endif
