@@ -1,6 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
-/* nm-openvpn-service-openvpn-helper - helper called after OpenVPN established
- * a connection, uses DBUS to send information back to nm-openvpn-service
+/*
+ * network-manager-openvpn - OpenVPN integration with NetworkManager
  *
  * Tim Niemueller [www.niemueller.de]
  * Based on work by Dan Williams <dcbw@redhat.com>
@@ -19,13 +18,11 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * (C) Copyright 2005 Red Hat, Inc.
- * (C) Copyright 2005 Tim Niemueller
- *
- * $Id: nm-openvpn-service-openvpn-helper.c 4170 2008-10-11 14:44:45Z dcbw $
+ * Copyright (C) 2005 Tim Niemueller
+ * Copyright (C) 2005 - 2018 Red Hat, Inc.
  */
 
-#include "config.h"
+#include "nm-default.h"
 
 #include <stdlib.h>
 #include <unistd.h>
@@ -38,29 +35,56 @@
 #include <arpa/inet.h>
 #include <ctype.h>
 #include <netdb.h>
+#include <syslog.h>
 
-#include "nm-default.h"
-
-#include "nm-openvpn-service.h"
-#include "nm-utils.h"
+#include "nm-utils/nm-shared-utils.h"
+#include "nm-utils/nm-vpn-plugin-macros.h"
 
 extern char **environ;
 
-static gboolean helper_debug = FALSE;
+static struct {
+	int log_level;
+	const char *log_prefix_token;
+} gl;
+
+/*****************************************************************************/
+
+#define _NMLOG(level, ...) \
+	G_STMT_START { \
+		if (gl.log_level >= (level)) { \
+			g_print ("nm-openvpn[%s]: %-7s [helper-%ld] " _NM_UTILS_MACRO_FIRST (__VA_ARGS__) "\n", \
+			         gl.log_prefix_token ?: "???", \
+			         nm_utils_syslog_to_str (level), \
+			         (long) getpid () \
+			         _NM_UTILS_MACRO_REST (__VA_ARGS__)); \
+		} \
+	} G_STMT_END
+
+static inline gboolean
+_LOGD_enabled (void)
+{
+	return gl.log_level >= LOG_INFO;
+}
+
+#define _LOGD(...) _NMLOG(LOG_INFO,    __VA_ARGS__)
+#define _LOGI(...) _NMLOG(LOG_NOTICE,  __VA_ARGS__)
+#define _LOGW(...) _NMLOG(LOG_WARNING, __VA_ARGS__)
+
+/*****************************************************************************/
 
 static void
 helper_failed (GDBusProxy *proxy, const char *reason)
 {
 	GError *err = NULL;
 
-	g_warning ("nm-openvpn-service-openvpn-helper did not receive a valid %s from openvpn", reason);
+	_LOGW ("nm-openvpn-service-openvpn-helper did not receive a valid %s from openvpn", reason);
 
 	if (!g_dbus_proxy_call_sync (proxy, "SetFailure",
 	                             g_variant_new ("(s)", reason),
 	                             G_DBUS_CALL_FLAGS_NONE, -1,
 	                             NULL,
 	                             &err)) {
-		g_warning ("Could not send failure information: %s", err->message);
+		_LOGW ("Could not send failure information: %s", err->message);
 		g_error_free (err);
 	}
 
@@ -78,7 +102,7 @@ send_config (GDBusProxy *proxy, GVariant *config,
 	                             G_DBUS_CALL_FLAGS_NONE, -1,
 	                             NULL,
 	                             &err)) {
-		g_warning ("Could not send configuration information: %s", err->message);
+		_LOGW ("Could not send configuration information: %s", err->message);
 		g_error_free (err);
 		err = NULL;
 	}
@@ -89,7 +113,7 @@ send_config (GDBusProxy *proxy, GVariant *config,
 	                                     G_DBUS_CALL_FLAGS_NONE, -1,
 	                                     NULL,
 		                             &err)) {
-			g_warning ("Could not send IPv4 configuration information: %s", err->message);
+			_LOGW ("Could not send IPv4 configuration information: %s", err->message);
 			g_error_free (err);
 			err = NULL;
 		}
@@ -101,7 +125,7 @@ send_config (GDBusProxy *proxy, GVariant *config,
 	                                     G_DBUS_CALL_FLAGS_NONE, -1,
 	                                     NULL,
 		                             &err)) {
-			g_warning ("Could not send IPv6 configuration information: %s", err->message);
+			_LOGW ("Could not send IPv6 configuration information: %s", err->message);
 			g_error_free (err);
 			err = NULL;
 		}
@@ -142,28 +166,6 @@ addr4_to_gvariant (const char *str)
 	return g_variant_new_uint32 (temp_addr.s_addr);
 }
 
-static void
-parse_addr4_list (GPtrArray *array, const char *str)
-{
-	char **split;
-	int i;
-	struct in_addr temp_addr;
-
-	/* Empty */
-	if (!str || strlen (str) < 1)
-		return;
-
-	split = g_strsplit (str, " ", -1);
-	for (i = 0; split[i]; i++) {
-		if (inet_pton (AF_INET, split[i], &temp_addr) > 0)
-			g_ptr_array_add (array, g_variant_new_uint32 (temp_addr.s_addr));
-	}
-
-	g_strfreev (split);
-
-	return;
-}
-
 static GVariant *
 addr6_to_gvariant (const char *str)
 {
@@ -184,151 +186,159 @@ addr6_to_gvariant (const char *str)
 	return g_variant_builder_end (&builder);
 }
 
+static void
+parse_addr_list (GPtrArray *array4, GPtrArray *array6, const char *str)
+{
+	char **split;
+	int i;
+	GVariant *variant;
+
+	/* Empty */
+	if (!str || strlen (str) < 1)
+		return;
+
+	split = g_strsplit (str, " ", -1);
+	for (i = 0; split[i]; i++) {
+		if (array4 && (variant = addr4_to_gvariant (split[i])) != NULL)
+			g_ptr_array_add (array4, variant);
+		else if (array6 && (variant = addr6_to_gvariant (split[i])) != NULL)
+			g_ptr_array_add (array6, variant);
+	}
+
+	g_strfreev (split);
+
+	return;
+}
+
 static inline gboolean
 is_domain_valid (const char *str)
 {
-	return (str && (strlen(str) >= 1) && (strlen(str) <= 255));
-}
+	gsize l;
 
-#define BUFLEN 256
+	if (!str)
+		return FALSE;
+	l = strlen (str);
+	return (l >= 1 && l <= 255);
+}
 
 static GVariant *
 get_ip4_routes (void)
 {
 	GVariantBuilder builder;
-	GVariant *value;
 	char *tmp;
-	int i, size = 0;
+	gboolean has_any = FALSE;
+	guint i;
 
 	g_variant_builder_init (&builder, G_VARIANT_TYPE ("aau"));
 
-	for (i = 1; i < 256; i++) {
+	for (i = 1;; i++) {
 		GVariantBuilder array;
-		char buf[BUFLEN];
-		struct in_addr network;
-		struct in_addr netmask;
-		struct in_addr gateway = { 0, };
-		guint32 prefix, metric = 0;
+		char key_name[255];
+		in_addr_t network;
+		in_addr_t netmask;
+		in_addr_t gateway = 0;
+		guint32 metric;
 
-		snprintf (buf, BUFLEN, "route_network_%d", i);
-		tmp = getenv (buf);
-		if (!tmp || strlen (tmp) < 1)
+		nm_sprintf_buf (key_name, "route_network_%u", i);
+		tmp = getenv (key_name);
+		if (!tmp || !tmp[0])
 			break;
 
-		if (inet_pton (AF_INET, tmp, &network) <= 0) {
-			g_warning ("Ignoring invalid static route address '%s'", tmp ? tmp : "NULL");
+		if (inet_pton (AF_INET, tmp, &network) != 1) {
+			_LOGW ("Ignoring invalid static route address %s = \"%s\"", key_name, tmp);
 			continue;
 		}
 
-		snprintf (buf, BUFLEN, "route_netmask_%d", i);
-		tmp = getenv (buf);
-		if (!tmp || inet_pton (AF_INET, tmp, &netmask) <= 0) {
-			g_warning ("Ignoring invalid static route netmask '%s'", tmp ? tmp : "NULL");
+		nm_sprintf_buf (key_name, "route_netmask_%u", i);
+		tmp = getenv (key_name);
+		if (   !tmp
+		    || inet_pton (AF_INET, tmp, &netmask) != 1) {
+			_LOGW ("Ignoring invalid static route netmask %s = %s%s%s", key_name, NM_PRINT_FMT_QUOTE_STRING (tmp));
 			continue;
 		}
 
-		snprintf (buf, BUFLEN, "route_gateway_%d", i);
-		tmp = getenv (buf);
+		nm_sprintf_buf (key_name, "route_gateway_%u", i);
+		tmp = getenv (key_name);
 		/* gateway can be missing */
-		if (tmp && (inet_pton (AF_INET, tmp, &gateway) <= 0)) {
-			g_warning ("Ignoring invalid static route gateway '%s'", tmp ? tmp : "NULL");
+		if (   tmp
+		    && inet_pton (AF_INET, tmp, &gateway) != 1) {
+			_LOGW ("Ignoring invalid static route gateway %s = \"%s\"", key_name, tmp);
 			continue;
 		}
 
-		snprintf (buf, BUFLEN, "route_metric_%d", i);
-		tmp = getenv (buf);
+		nm_sprintf_buf (key_name, "route_metric_%u", i);
+		tmp = getenv (key_name);
 		/* metric can be missing */
-		if (tmp && strlen (tmp)) {
-			long int tmp_metric;
-
-			errno = 0;
-			tmp_metric = strtol (tmp, NULL, 10);
-			if (errno || tmp_metric < 0 || tmp_metric > G_MAXUINT32) {
-				g_warning ("Ignoring invalid static route metric '%s'", tmp);
+		if (tmp && tmp[0]) {
+			metric = _nm_utils_ascii_str_to_int64 (tmp, 10, 0, G_MAXUINT32, 0);
+			if (errno) {
+				_LOGW ("Ignoring invalid static route metric %s = \"%s\"", key_name, tmp);
 				continue;
 			}
-			metric = (guint32) tmp_metric;
-		}
+		} else
+			metric = 0;
 
 		g_variant_builder_init (&array, G_VARIANT_TYPE ("au"));
-		g_variant_builder_add_value (&array, g_variant_new_uint32 (network.s_addr));
-		prefix = nm_utils_ip4_netmask_to_prefix (netmask.s_addr);
-		g_variant_builder_add_value (&array, g_variant_new_uint32 (prefix));
-		g_variant_builder_add_value (&array, g_variant_new_uint32 (gateway.s_addr));
+		g_variant_builder_add_value (&array, g_variant_new_uint32 (network));
+		g_variant_builder_add_value (&array, g_variant_new_uint32 (nm_utils_ip4_netmask_to_prefix (netmask)));
+		g_variant_builder_add_value (&array, g_variant_new_uint32 (gateway));
 		g_variant_builder_add_value (&array, g_variant_new_uint32 (metric));
 		g_variant_builder_add_value (&builder, g_variant_builder_end (&array));
-		size++;
+		has_any = TRUE;
 	}
 
-	value = g_variant_builder_end (&builder);
-	if (size > 0)
-		return value;
+	if (!has_any) {
+		g_variant_builder_clear (&builder);
+		return NULL;
+	}
 
-	g_variant_unref (value);
-	return NULL;
+	return g_variant_builder_end (&builder);
 }
 
 static GVariant *
-get_ip6_routes (void)
+get_ip6_routes (const char *fallback_gateway)
 {
-	GVariant *value = NULL;
-	GPtrArray *routes;
-	char *tmp;
-	int i;
+	gs_unref_ptrarray GPtrArray *routes = NULL;
+	guint i;
 
-	routes = g_ptr_array_new_full (256, (GDestroyNotify) nm_ip_route_unref);
+	routes = g_ptr_array_new_full (10, (GDestroyNotify) nm_ip_route_unref);
 
-	for (i = 1; i < 256; i++) {
+	for (i = 1;; i++) {
 		NMIPRoute *route;
-		char buf[BUFLEN];
-		guint32 prefix;
-		gchar **dest_prefix;
 		GError *error = NULL;
+		gs_free char *dst = NULL;
+		char key_name[255];
+		int prefix;
+		const char *tmp;
 
-		snprintf (buf, BUFLEN, "route_ipv6_network_%d", i);
-		tmp = getenv (buf);
-		if (!tmp || strlen (tmp) < 1)
+		nm_sprintf_buf (key_name, "route_ipv6_network_%u", i);
+		tmp = getenv (key_name);
+		if (!tmp || !tmp[0])
 			break;
 
-		/* Split network string in "dest/prefix" format */
-		dest_prefix = g_strsplit (tmp, "/", 2);
-
-		tmp = dest_prefix[1];
-		if (tmp) {
-			long int tmp_prefix;
-
-			errno = 0;
-			tmp_prefix = strtol (tmp, NULL, 10);
-			if (errno || tmp_prefix <= 0 || tmp_prefix > 128) {
-				g_warning ("Ignoring invalid static route prefix '%s'", tmp ? tmp : "NULL");
-				g_strfreev (dest_prefix);
-				continue;
-			}
-			prefix = (guint32) tmp_prefix;
-		} else {
-			g_warning ("Ignoring static route %d with no prefix length", i);
-			g_strfreev (dest_prefix);
+		if (   !nm_utils_parse_inaddr_prefix (AF_INET6, tmp, &dst, &prefix)
+		    || prefix == -1) {
+			_LOGW ("Ignoring invalid static route %s = \"%s\"", key_name, tmp);
 			continue;
 		}
 
-		snprintf (buf, BUFLEN, "route_ipv6_gateway_%d", i);
-		tmp = getenv (buf);
+		nm_sprintf_buf (key_name, "route_ipv6_gateway_%u", i);
+		tmp = getenv (key_name);
 
-		route = nm_ip_route_new (AF_INET6, dest_prefix[0], prefix, tmp, -1, &error);
-		g_strfreev (dest_prefix);
+		route = nm_ip_route_new (AF_INET6, dst, prefix, tmp ?: fallback_gateway, -1, &error);
 		if (!route) {
-			g_warning ("Ignoring a route: %s", error->message);
+			_LOGW ("Ignoring route#%u: %s", i, error->message);
 			g_error_free (error);
 			continue;
 		}
+
 		g_ptr_array_add (routes, route);
 	}
 
-	if (routes->len)
-		value = nm_utils_ip6_routes_to_variant (routes);
-	g_ptr_array_unref (routes);
+	if (!routes->len)
+		return NULL;
 
-	return value;
+	return nm_utils_ip6_routes_to_variant (routes);
 }
 
 static GVariant *
@@ -343,8 +353,8 @@ trusted_remote_to_gvariant (void)
 	if (tmp) {
 		val = addr6_to_gvariant (tmp);
 		if (val == NULL) {
-			g_warning ("%s: failed to convert VPN gateway address '%s' (%d)",
-			           __func__, tmp, errno);
+			_LOGW ("failed to convert VPN gateway address '%s' (%d)",
+			       tmp, errno);
 			return NULL;
 		}
 		return val;
@@ -354,7 +364,7 @@ trusted_remote_to_gvariant (void)
 	if (!tmp)
 		tmp = getenv ("remote_1");
 	if (!tmp) {
-		g_warning ("%s: did not receive remote gateway address", __func__);
+		_LOGW ("did not receive remote gateway address");
 		return NULL;
 	}
 
@@ -382,8 +392,8 @@ trusted_remote_to_gvariant (void)
 		hints.ai_flags = AI_ADDRCONFIG;
 		err = getaddrinfo (tmp, NULL, &hints, &result);
 		if (err != 0) {
-			g_warning ("%s: failed to look up VPN gateway address '%s' (%d)",
-			           __func__, tmp, err);
+			_LOGW ("failed to look up VPN gateway address '%s' (%d)",
+			       tmp, err);
 			return NULL;
 		}
 
@@ -405,15 +415,15 @@ trusted_remote_to_gvariant (void)
 		if (addr.s_addr != 0)
 			return g_variant_new_uint32 (addr.s_addr);
 		else {
-			g_warning ("%s: failed to convert or look up VPN gateway address '%s'",
-			           __func__, tmp);
+			_LOGW ("failed to convert or look up VPN gateway address '%s'",
+			       tmp);
 			return NULL;
 		}
 	} else {
 		val = addr4_to_gvariant (tmp);
 		if (val == NULL) {
-			g_warning ("%s: failed to convert VPN gateway address '%s' (%d)",
-			           __func__, tmp, errno);
+			_LOGW ("failed to convert VPN gateway address '%s' (%d)",
+			       tmp, errno);
 			return NULL;
 		}
 	}
@@ -431,7 +441,7 @@ main (int argc, char *argv[])
 	GVariant *val;
 	int i;
 	GError *err = NULL;
-	GPtrArray *dns_list;
+	GPtrArray *dns4_list, *dns6_list;
 	GPtrArray *nbns_list;
 	GPtrArray *dns_domains;
 	struct in_addr temp_addr;
@@ -441,7 +451,9 @@ main (int argc, char *argv[])
 	gboolean is_restart;
 	gboolean has_ip4_prefix = FALSE;
 	gboolean has_ip4_address = FALSE;
+	gboolean has_ip6_address = FALSE;
 	gchar *bus_name = NM_DBUS_SERVICE_OPENVPN;
+	gsize size;
 
 #if !GLIB_CHECK_VERSION (2, 35, 0)
 	g_type_init ();
@@ -452,19 +464,24 @@ main (int argc, char *argv[])
 			i++;
 			break;
 		}
-		if (!strcmp (argv[i], "--helper-debug"))
-			helper_debug = TRUE;
-		else if (!strcmp (argv[i], "--tun"))
+		if (nm_streq (argv[i], "--debug")) {
+			if (i + 2 >= argc) {
+				g_printerr ("Missing debug arguments (requires <LEVEL> <PREFIX_TOKEN>)\n");
+				exit (1);
+			}
+			gl.log_level = _nm_utils_ascii_str_to_int64 (argv[++i], 10, 0, LOG_DEBUG, 0);
+			gl.log_prefix_token = argv[++i];
+		} else if (!strcmp (argv[i], "--tun"))
 			tapdev = 0;
 		else if (!strcmp (argv[i], "--tap"))
 			tapdev = 1;
 		else if (!strcmp (argv[i], "--bus-name")) {
 			if (++i == argc) {
-				g_warning ("Missing bus name argument");
+				g_printerr ("Missing bus name argument\n");
 				exit (1);
 			}
 			if (!g_dbus_is_name (argv[i])) {
-				g_warning ("Invalid bus name");
+				g_printerr ("Invalid bus name\n");
 				exit (1);
 			}
 			bus_name = argv[i];
@@ -473,7 +490,7 @@ main (int argc, char *argv[])
 	}
 	shift = i - 1;
 
-	if (helper_debug) {
+	if (_LOGD_enabled ()) {
 		GString *args;
 
 		args = g_string_new (NULL);
@@ -487,13 +504,11 @@ main (int argc, char *argv[])
 			g_free (tmp);
 		}
 
-		g_message ("command line: %s", args->str);
+		_LOGD ("command line: %s", args->str);
 		g_string_free (args, TRUE);
-		g_message ("openvpn script environment ---------------------------");
-		iter = environ;
-		while (iter && *iter)
-			g_message ("%s", *iter++);
-		g_message ("------------------------------------------------------");
+
+		for (iter = environ; iter && *iter; iter++)
+			_LOGD ("environment: %s", *iter);
 	}
 
 	/* shift the arguments to the right leaving only those provided by openvpn */
@@ -504,14 +519,14 @@ main (int argc, char *argv[])
 	is_restart = argc >= 7 && !g_strcmp0 (argv[6], "restart");
 
 	proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
-	                                       G_DBUS_PROXY_FLAGS_NONE,
+	                                       G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
 	                                       NULL,
 	                                       bus_name,
 	                                       NM_VPN_DBUS_PLUGIN_PATH,
 	                                       NM_VPN_DBUS_PLUGIN_INTERFACE,
 	                                       NULL, &err);
 	if (!proxy) {
-		g_warning ("Could not create a D-Bus proxy: %s", err->message);
+		_LOGW ("Could not create a D-Bus proxy: %s", err->message);
 		g_error_free (err);
 		exit (1);
 	}
@@ -538,6 +553,10 @@ main (int argc, char *argv[])
 			g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_INT_GATEWAY, val);
 	}
 
+	/* tell Network Manager that openvpn can persist */
+	val = g_variant_new_boolean (TRUE);
+	g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CAN_PERSIST, val);
+
 	/* VPN device */
 	tmp = getenv ("dev");
 	val = str_to_gvariant (tmp, FALSE);
@@ -602,32 +621,28 @@ main (int argc, char *argv[])
 			g_variant_builder_add (&ip4builder, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_PREFIX, val);
 		}
 	} else
-		g_warning ("No IP4 netmask/prefix (missing or invalid 'ifconfig_netmask')");
+		_LOGW ("No IP4 netmask/prefix (missing or invalid 'ifconfig_netmask')");
 
 	val = get_ip4_routes ();
 	if (val)
 		g_variant_builder_add (&ip4builder, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_ROUTES, val);
+	else if (is_restart) {
+		g_variant_builder_add (&ip4builder, "{sv}",
+		                       NM_VPN_PLUGIN_IP4_CONFIG_PRESERVE_ROUTES,
+		                       g_variant_new_boolean (TRUE));
+	}
 
 	/* IPv6 address */
 	tmp = getenv ("ifconfig_ipv6_local");
 	if (tmp && strlen (tmp)) {
 		val = addr6_to_gvariant (tmp);
-		if (val)
+		if (val) {
 			g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_ADDRESS, val);
-		else
+			has_ip6_address = TRUE;
+		} else
 			helper_failed (proxy, "IP6 Address");
 	}
 
-	/* IPv6 remote address */
-	tmp = getenv ("ifconfig_ipv6_remote");
-	if (tmp && strlen (tmp)) {
-		val = addr6_to_gvariant (tmp);
-		if (val)
-			g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_PTP, val);
-		else
-			helper_failed (proxy, "IP6 PTP Address");
-	}
-
 	/* IPv6 netbits */
 	tmp = getenv ("ifconfig_ipv6_netbits");
 	if (tmp && strlen (tmp)) {
@@ -636,50 +651,66 @@ main (int argc, char *argv[])
 		errno = 0;
 		netbits = strtol (tmp, NULL, 10);
 		if (errno || netbits < 0 || netbits > 128) {
-			g_warning ("Ignoring invalid prefix '%s'", tmp);
+			_LOGW ("Ignoring invalid prefix '%s'", tmp);
 		} else {
 			val = g_variant_new_uint32 ((guint32) netbits);
 			g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_PREFIX, val);
 		}
 	}
 
-	val = get_ip6_routes ();
+	/* Note: for IPv6 'ifconfig_ipv6_remote' is not used as the peer
+	 * address but as fallback gateway for routes.
+	 */
+	val = get_ip6_routes (getenv ("ifconfig_ipv6_remote"));
 	if (val)
 		g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_ROUTES, val);
+	else if (is_restart) {
+		g_variant_builder_add (&ip6builder, "{sv}",
+		                       NM_VPN_PLUGIN_IP6_CONFIG_PRESERVE_ROUTES,
+		                       g_variant_new_boolean (TRUE));
+	}
 
 	/* DNS and WINS servers */
 	dns_domains = g_ptr_array_sized_new (3);
-	dns_list = g_ptr_array_new ();
+	dns4_list = g_ptr_array_new ();
+	dns6_list = g_ptr_array_new ();
 	nbns_list = g_ptr_array_new ();
 
-	for (i = 1; i < 256; i++) {
-		char *env_name;
+	for (i = 1; TRUE; i++) {
+		char env_name[NM_STRLEN ("foreign_option_") + 1 + 40];
+
+		nm_sprintf_buf (env_name, "foreign_option_%d", i);
 
-		env_name = g_strdup_printf ("foreign_option_%d", i);
 		tmp = getenv (env_name);
-		g_free (env_name);
 
-		if (!tmp || strlen (tmp) < 1)
+		if (   !tmp
+		    || !tmp[0])
 			break;
 
 		if (!g_str_has_prefix (tmp, "dhcp-option "))
 			continue;
 
-		tmp += 12; /* strlen ("dhcp-option ") */
+		tmp += NM_STRLEN ("dhcp-option ");
 
 		if (g_str_has_prefix (tmp, "DNS "))
-			parse_addr4_list (dns_list, tmp + 4);
+			parse_addr_list (dns4_list, dns6_list, &tmp[NM_STRLEN ("DNS ")]);
 		else if (g_str_has_prefix (tmp, "WINS "))
-			parse_addr4_list (nbns_list, tmp + 5);
-		else if (g_str_has_prefix (tmp, "DOMAIN ") && is_domain_valid (tmp + 7))
-			g_ptr_array_add (dns_domains, tmp + 7);
+			parse_addr_list (nbns_list, NULL, &tmp[NM_STRLEN ("WINS ")]);
+		else if (   g_str_has_prefix (tmp, "DOMAIN ")
+		         && is_domain_valid (&tmp[NM_STRLEN ("DOMAIN ")]))
+			g_ptr_array_add (dns_domains, &tmp[NM_STRLEN ("DOMAIN ")]);
 	}
 
-	if (dns_list->len) {
-		val = g_variant_new_array (G_VARIANT_TYPE_UINT32, (GVariant **) dns_list->pdata, dns_list->len);
+	if (dns4_list->len) {
+		val = g_variant_new_array (G_VARIANT_TYPE_UINT32, (GVariant **) dns4_list->pdata, dns4_list->len);
 		g_variant_builder_add (&ip4builder, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_DNS, val);
 	}
 
+	if (has_ip6_address && dns6_list->len) {
+		val = g_variant_new_array (G_VARIANT_TYPE ("ay"), (GVariant **) dns6_list->pdata, dns6_list->len);
+		g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_DNS, val);
+	}
+
 	if (nbns_list->len) {
 		val = g_variant_new_array (G_VARIANT_TYPE_UINT32, (GVariant **) nbns_list->pdata, nbns_list->len);
 		g_variant_builder_add (&ip4builder, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_NBNS, val);
@@ -688,45 +719,58 @@ main (int argc, char *argv[])
 	if (dns_domains->len) {
 		val = g_variant_new_strv ((const gchar **) dns_domains->pdata, dns_domains->len);
 		g_variant_builder_add (&ip4builder, "{sv}", NM_VPN_PLUGIN_IP4_CONFIG_DOMAINS, val);
+
+		/* Domains apply to both IPv4 and IPv6 configurations */
+		if (has_ip6_address) {
+			val = g_variant_new_strv ((const gchar **) dns_domains->pdata, dns_domains->len);
+			g_variant_builder_add (&ip6builder, "{sv}", NM_VPN_PLUGIN_IP6_CONFIG_DOMAINS, val);
+		}
 	}
 
-	g_ptr_array_unref (dns_list);
+	g_ptr_array_unref (dns4_list);
+	g_ptr_array_unref (dns6_list);
 	g_ptr_array_unref (nbns_list);
 	g_ptr_array_unref (dns_domains);
 
 	/* Tunnel MTU */
 	tmp = getenv ("tun_mtu");
-	if (tmp && strlen (tmp)) {
-		long int mtu;
+	if (tmp && tmp[0]) {
+		guint32 mtu;
 
-		errno = 0;
-		mtu = strtol (tmp, NULL, 10);
-		if (errno || mtu < 0 || mtu > 20000) {
-			g_warning ("Ignoring invalid tunnel MTU '%s'", tmp);
-		} else {
-			val = g_variant_new_uint32 ((guint32) mtu);
-			g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CONFIG_MTU, val);
+		mtu = _nm_utils_ascii_str_to_int64 (tmp, 10, 0, G_MAXUINT32, 0);
+		if (errno)
+			_LOGW ("Ignoring invalid tunnel MTU '%s'", tmp);
+		else {
+			g_variant_builder_add (&builder, "{sv}",
+			                       NM_VPN_PLUGIN_CONFIG_MTU,
+			                       g_variant_new_uint32 (mtu));
 		}
 	}
 
 	ip4config = g_variant_builder_end (&ip4builder);
 
-	if (g_variant_n_children (ip4config)) {
-		val = g_variant_new_boolean (TRUE);
-		g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CONFIG_HAS_IP4, val);
-	} else {
+	size = g_variant_n_children (ip4config);
+	if (size == 0 || !has_ip4_address) {
+		if (size > 0)
+			_LOGW ("Ignoring IPv4 configuration without an address");
 		g_variant_unref (ip4config);
 		ip4config = NULL;
+	} else {
+		val = g_variant_new_boolean (TRUE);
+		g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CONFIG_HAS_IP4, val);
 	}
 
 	ip6config = g_variant_builder_end (&ip6builder);
 
-	if (g_variant_n_children (ip6config)) {
-		val = g_variant_new_boolean (TRUE);
-		g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CONFIG_HAS_IP6, val);
-	} else {
+	size = g_variant_n_children (ip6config);
+	if (size == 0 || !has_ip6_address) {
+		if (size > 0)
+			_LOGW ("Ignoring IPv6 configuration without an address");
 		g_variant_unref (ip6config);
 		ip6config = NULL;
+	} else {
+		val = g_variant_new_boolean (TRUE);
+		g_variant_builder_add (&builder, "{sv}", NM_VPN_PLUGIN_CONFIG_HAS_IP6, val);
 	}
 
 	if (!ip4config && !ip6config)
