@@ -1,8 +1,5 @@
-/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
-/* nm-openvpn-service - openvpn integration with NetworkManager
- *
- * Copyright (C) 2005 - 2008 Tim Niemueller <tim@niemueller.de>
- * Copyright (C) 2005 - 2010 Dan Williams <dcbw@redhat.com>
+/*
+ * network-manager-openvpn - OpenVPN integration with NetworkManager
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,8 +15,9 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * $Id: nm-openvpn-service.c 4232 2008-10-29 09:13:40Z tambeti $
- *
+ * Copyright (C) 2005 - 2008 Tim Niemueller <tim@niemueller.de>
+ * Copyright (C) 2005 - 2010 Dan Williams <dcbw@redhat.com>
+ * Copyright (C) 2008 - 2018 Red Hat, Inc.
  */
 
 #include "nm-default.h"
@@ -140,11 +138,13 @@ typedef struct {
 } ValidProperty;
 
 static const ValidProperty valid_properties[] = {
+	{ NM_OPENVPN_KEY_ALLOW_PULL_FQDN,           G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_AUTH,                      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CA,                        G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CERT,                      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CIPHER,                    G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_KEYSIZE,                   G_TYPE_INT, 1, 65535, FALSE },
+	{ NM_OPENVPN_KEY_COMPRESS,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_COMP_LZO,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CONNECT_TIMEOUT,           G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_CONNECTION_TYPE,           G_TYPE_STRING, 0, 0, FALSE },
@@ -152,6 +152,7 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CRL_VERIFY_DIR,            G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_EXTRA_CERTS,               G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FLOAT,                     G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_NCP_DISABLE,               G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FRAGMENT_SIZE,             G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_KEY,                       G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_LOCAL_IP,                  G_TYPE_STRING, 0, 0, TRUE },
@@ -167,9 +168,11 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_PROXY_SERVER,              G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_PROXY_PORT,                G_TYPE_INT, 1, 65535, FALSE },
 	{ NM_OPENVPN_KEY_PROXY_RETRY,               G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_PUSH_PEER_INFO,            G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_HTTP_PROXY_USERNAME,       G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE,                    G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE_RANDOM,             G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME,    G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE_IP,                 G_TYPE_STRING, 0, 0, TRUE },
 	{ NM_OPENVPN_KEY_RENEG_SECONDS,             G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_STATIC_KEY,                G_TYPE_STRING, 0, 0, FALSE },
@@ -182,6 +185,7 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_TUN_IPV6,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_CIPHER,                G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_CRYPT,                 G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_CRYPT_V2,              G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_REMOTE,                G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_VERIFY_X509_NAME,          G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE_CERT_TLS,           G_TYPE_STRING, 0, 0, FALSE },
@@ -192,6 +196,8 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CERTPASS_FLAGS,            G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_NOSECRET,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD_FLAGS, G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_VERSION_MIN,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_VERSION_MAX,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                                     G_TYPE_NONE, FALSE }
 };
 
@@ -352,18 +358,14 @@ args_add_vpn_certs (GPtrArray *args, NMSettingVpn *s_vpn)
 	key  = nm_utils_str_utf8safe_unescape (key,  &key_free);
 
 	if (   nmovpn_arg_is_set (ca)
-	    && nmovpn_arg_is_set (cert)
-	    && nmovpn_arg_is_set (key)
-	    && nm_streq (ca, cert)
-	    && nm_streq (ca, key))
-		args_add_strv (args, "--pkcs12", ca);
+	    && !is_pkcs12 (ca))
+		args_add_strv (args, "--ca", ca);
+
+	if (nmovpn_arg_is_set (cert) && is_pkcs12 (cert))
+		args_add_strv (args, "--pkcs12", cert);
 	else {
-		if (nmovpn_arg_is_set (ca))
-			args_add_strv (args, "--ca", ca);
-		if (nmovpn_arg_is_set (cert))
-			args_add_strv (args, "--cert", cert);
-		if (nmovpn_arg_is_set (key))
-			args_add_strv (args, "--key", key);
+		args_add_strv (args, "--cert", cert);
+		args_add_strv (args, "--key", key);
 	}
 }
 
@@ -546,8 +548,9 @@ openvpn_binary_detect_version (const char *exepath)
 		return OPENVPN_BINARY_VERSION_UNKNOWN;
 
 	if (   !WIFEXITED (exit_code)
-	    || WEXITSTATUS (exit_code) != 1) {
-		/* expect return code 1 (OPENVPN_EXIT_STATUS_USAGE) */
+	    || !NM_IN_SET(WEXITSTATUS (exit_code), 0, 1)) {
+		/* expect return code 1 (OPENVPN_EXIT_STATUS_USAGE).
+		 * Since 2.5.0, it returns 0. */
 		return OPENVPN_BINARY_VERSION_UNKNOWN;
 	}
 
@@ -576,8 +579,23 @@ openvpn_binary_detect_version (const char *exepath)
 static OpenvpnBinaryVersion
 openvpn_binary_detect_version_cached (const char *exepath, OpenvpnBinaryVersion *cached)
 {
-	if (G_UNLIKELY (*cached == OPENVPN_BINARY_VERSION_INVALID))
+	if (G_UNLIKELY (*cached == OPENVPN_BINARY_VERSION_INVALID)) {
+		const char *str;
+
 		*cached = openvpn_binary_detect_version (exepath);
+		switch (*cached) {
+		case OPENVPN_BINARY_VERSION_2_3_OR_OLDER:
+			str = "2.3 or older";
+			break;
+		case OPENVPN_BINARY_VERSION_2_4_OR_NEWER:
+			str = "2.4 or newer";
+			break;
+		default:
+			str = "unknown";
+			break;
+		}
+		_LOGI ("detected openvpn version %s", str);
+	}
 	return *cached;
 }
 
@@ -1318,6 +1336,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	GPid pid;
 	gboolean dev_type_is_tap;
 	const char *defport, *proto_tcp;
+	const char *compress;
 	const char *tls_remote = NULL;
 	const char *nm_openvpn_user, *nm_openvpn_group, *nm_openvpn_chroot;
 	gs_free char *bus_name = NULL;
@@ -1327,6 +1346,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	OpenvpnBinaryVersion openvpn_binary_version = OPENVPN_BINARY_VERSION_INVALID;
 	guint num_remotes = 0;
 	gs_free char *cmd_log = NULL;
+	NMOvpnComp comp;
 
 	s_vpn = nm_connection_get_setting_vpn (connection);
 	if (!s_vpn) {
@@ -1433,8 +1453,10 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 				}
 			} else if (nm_streq0 (proto_tcp, "yes"))
 				args_add_strv (args, "tcp-client");
-			else
+			else {
 				args_add_strv (args, "udp");
+				args_add_strv (args, "--explicit-exit-notify");
+			}
 		}
 	}
 
@@ -1442,6 +1464,14 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	if (nm_streq0 (tmp, "yes"))
 		args_add_strv (args, "--remote-random");
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_RANDOM_HOSTNAME);
+	if (nm_streq0 (tmp, "yes"))
+		args_add_strv (args, "--remote-random-hostname");
+
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_ALLOW_PULL_FQDN);
+	if (nm_streq0 (tmp, "yes"))
+		args_add_strv (args, "--allow-pull-fqdn");
+
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TUN_IPV6);
 	if (nm_streq0 (tmp, "yes"))
 		args_add_strv (args, "--tun-ipv6");
@@ -1473,8 +1503,6 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		}
 	}
 
-	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
-
 	/* openvpn understands 4 different modes for --comp-lzo, which have
 	 * different meaning:
 	 *  1) no --comp-lzo option
@@ -1492,19 +1520,60 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	 *
 	 * See bgo#769177
 	 */
-	if (NM_IN_STRSET (tmp, "no")) {
-		/* means no --comp-lzo option. */
-		tmp = NULL;
-	} else if (NM_IN_STRSET (tmp, "no-by-default"))
-		tmp = "no";
 
-	if (NM_IN_STRSET (tmp, "yes", "no", "adaptive"))
-		args_add_strv (args, "--comp-lzo", tmp);
+	/* New (2.4+) compress option ("lz4", "lzo", ...) */
+	compress = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMPRESS);
+	/* Legacy option ("yes", "adaptive", "no", ...) */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
+
+	if (compress && tmp)
+		_LOGW ("'compress' option overrides 'comp-lzo'");
+
+	comp = nmovpn_compression_from_options (tmp, compress);
+	openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version);
+
+	switch (comp) {
+	case NMOVPN_COMP_DISABLED:
+		break;
+	case NMOVPN_COMP_LZO:
+		if (openvpn_binary_version == OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
+			args_add_strv (args, "--comp-lzo", "yes");
+		else
+			args_add_strv (args, "--compress", "lzo");
+		break;
+	case NMOVPN_COMP_LZ4:
+	case NMOVPN_COMP_LZ4_V2:
+	case NMOVPN_COMP_AUTO:
+		if (openvpn_binary_version == OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
+			_LOGW ("\"compress\" option supported only by OpenVPN >= 2.4");
+
+		if (comp == NMOVPN_COMP_LZ4)
+			args_add_strv (args, "--compress", "lz4");
+		else if (comp == NMOVPN_COMP_LZ4_V2)
+			args_add_strv (args, "--compress", "lz4-v2");
+		else
+			args_add_strv (args, "--compress");
+		break;
+	case NMOVPN_COMP_LEGACY_LZO_DISABLED:
+	case NMOVPN_COMP_LEGACY_LZO_ADAPTIVE:
+		if (openvpn_binary_version != OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
+			_LOGW ("\"comp-lzo\" is deprecated and will be removed in future OpenVPN releases");
+
+		args_add_strv (args, "--comp-lzo",
+		                  comp == NMOVPN_COMP_LEGACY_LZO_DISABLED
+		               ? "no"
+		               : "adaptive");
+		break;
+	}
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_FLOAT);
 	if (nm_streq0 (tmp, "yes"))
 		args_add_strv (args, "--float");
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_NCP_DISABLE);
+	if (nm_streq0 (tmp, "yes"))
+		args_add_strv (args, "--ncp-disable");
+
 	/* ping, ping-exit, ping-restart */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PING);
 	if (tmp) {
@@ -1648,6 +1717,22 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		args_add_utf8safe_str (args, tmp);
 	}
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_CRYPT_V2);
+	if (nmovpn_arg_is_set (tmp)) {
+		args_add_strv (args, "--tls-crypt-v2");
+		args_add_utf8safe_str (args, tmp);
+	}
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MIN);
+	if (nmovpn_arg_is_set (tmp)) {
+		args_add_strv (args, "--tls-version-min");
+		args_add_strv (args, tmp);
+	}
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MAX);
+	if (nmovpn_arg_is_set (tmp)) {
+		args_add_strv (args, "--tls-version-max");
+		args_add_strv (args, tmp);
+	}
+
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_EXTRA_CERTS);
 	if (nmovpn_arg_is_set (tmp)) {
 		args_add_strv (args, "--extra-certs");
@@ -1656,11 +1741,11 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_REMOTE);
 	if (nmovpn_arg_is_set (tmp)) {
-		if (openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version) != OPENVPN_BINARY_VERSION_2_4_OR_NEWER) {
-			_LOGW ("the tls-remote option is deprecated and removed from OpenVPN 2.4. Update your connection to use verify-x509-name");
+		if (openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version) == OPENVPN_BINARY_VERSION_2_3_OR_OLDER) {
+			_LOGW ("the tls-remote option is deprecated and removed from OpenVPN 2.4. Update your connection to use verify-x509-name (for example, \"verify-x509-name=name:%s\")", tmp);
 			args_add_strv (args, "--tls-remote", tmp);
 		} else {
-			_LOGW ("the tls-remote option is deprecated and removed from OpenVPN 2.4. For compatibility, the plugin uses \"verify-x509-name\" \"%s\" \"name\" instead. Update your connection to use verify-x509-name", tmp);
+			_LOGW ("the tls-remote option is deprecated and removed from OpenVPN 2.4. For compatibility, the plugin uses \"verify-x509-name\" \"%s\" \"name\" instead. Update your connection to use for example \"verify-x509-name=name:%s\")", tmp, tmp);
 			args_add_strv (args, "--verify-x509-name", tmp, "name");
 		}
 		tls_remote = tmp;
@@ -1763,7 +1848,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	if (tmp) {
 		if (nm_streq (tmp, "yes"))
 			args_add_strv (args, "--mssfix");
-		else if ((v_int64 = _nm_utils_ascii_str_to_int64 (tmp, 10, 1, G_MAXINT32, 0))) {
+		else if ((v_int64 = _nm_utils_ascii_str_to_int64 (tmp, 10, 0, G_MAXINT32, -1)) != -1) {
 			args_add_strv (args, "--mssfix");
 			args_add_int64 (args, v_int64);
 		}
@@ -1826,6 +1911,10 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	args_add_strv (args, "--route-noexec");
 	args_add_strv (args, "--ifconfig-noexec");
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PUSH_PEER_INFO);
+	if (nm_streq0 (tmp, "yes"))
+		args_add_strv (args, "--push-peer-info");
+
 	/* Now append configuration options which are dependent on the configuration type */
 	if (nm_streq (connection_type, NM_OPENVPN_CONTYPE_TLS)) {
 		args_add_strv (args, "--client");
